#include <winsock2.h>
#include <windows.h>
#include <process.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdlib.h>

#include "server.h"
#include "resource.h"
#include "user.h"
#include "world.h"
#include "protocol.h"

#define MAINWNDCLASS "_SERVERWND_"
#define WM_USER_LOG (WM_USER + 0)

LONG APIENTRY WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

HWND g_hwndMain = NULL;
HANDLE g_hMutex = NULL;
HFONT g_hArialFont = NULL;
HWND g_hwndLog = NULL;
SOCKET g_hListener = NULL;
HANDLE g_hServerThread = NULL;
int g_nPort = 1001;
std::set<SOCKET> g_setSocket;
CRITICAL_SECTION g_csSocket;
bool g_bShutDown = false;
World g_world;

unsigned __stdcall ServerThread(void* pArg);
unsigned __stdcall UserThread(void* pArg);

int PASCAL WinMain(HINSTANCE hInstance,HINSTANCE hPrev,LPSTR lpszCmdLine,int cmdShow)
{
    MSG	msg;
    WNDCLASS wc;

    wc.style = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS;
    wc.lpfnWndProc = WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_SERVER));
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) NULL;
    wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
    wc.lpszClassName = MAINWNDCLASS;

    if (RegisterClass((LPWNDCLASS) &wc) == false) {
        return(FALSE);
	}
    g_hwndMain = CreateWindow(
		MAINWNDCLASS,
        "Server",
        WS_OVERLAPPEDWINDOW | WS_VISIBLE,
		0, 0, 640, 480,
		NULL,
		NULL,
        hInstance,
        NULL);

    while (GetMessage(&msg, NULL, 0, 0) ) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    return msg.wParam;
}

LONG APIENTRY WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT	ps;
	HDC hDC;

    switch (message) {
    case WM_CREATE :
		g_hwndLog = CreateWindowEx(WS_EX_CLIENTEDGE, "listbox", "", 
			WS_CHILD | WS_VISIBLE | WS_VSCROLL | LBS_NOINTEGRALHEIGHT,
			0, 0, 0, 0,
			hWnd,
			(HMENU) 101,
			(HINSTANCE) GetWindowLong(hWnd, GWL_HINSTANCE),
			NULL);
		if (g_hwndLog != NULL) {
			LOGFONT logfont;

			memset(&logfont, 0, sizeof(LOGFONT));
			logfont.lfCharSet = DEFAULT_CHARSET;
			logfont.lfHeight = -12;
			strcpy(logfont.lfFaceName, "Arial");
			g_hArialFont = CreateFontIndirect(&logfont);
			SendMessage(g_hwndLog, WM_SETFONT, (WPARAM) g_hArialFont, (LPARAM) 0);
		}
		InitializeCriticalSection(&g_csSocket);
		g_hServerThread = (HANDLE) _beginthreadex(NULL, 0, ServerThread, 0, 0, NULL);
		break;
    case WM_CLOSE :
		DestroyWindow(hWnd);
		break;
	case WM_SIZE :
		if (wParam != SIZE_MINIMIZED) {
			MoveWindow(g_hwndLog, 0, 0, LOWORD(lParam), HIWORD(lParam), TRUE);
		}
		break;
	case WM_USER_LOG :
		SendMessage(g_hwndLog, LB_ADDSTRING, 0, lParam);
		break;
    case WM_DESTROY :
		if (g_hArialFont != NULL) {
			DeleteObject(g_hArialFont);
		}
		PostQuitMessage(0);
        break;
	case WM_COMMAND :
		if (wParam == ID_FILE_EXIT) {
			DestroyWindow(hWnd);
		}
		break;
    case WM_PAINT :
        hDC = BeginPaint(hWnd, &ps);
        EndPaint(hWnd, &ps);
        break;
    default :
        return DefWindowProc(hWnd, message, wParam, lParam);
    }
    return 0;
}

void Log(char* pszLog)
{
	SendMessage(g_hwndLog, LB_ADDSTRING, 0, (LPARAM) pszLog);
}

unsigned __stdcall ServerThread(void* pArg)
{
	char szBuffer[1024];
    WSADATA wd;

	// socket ʱȭ
	if (WSAStartup(0x0202, &wd) != 0) {
		sprintf(szBuffer, "WSAStartup error code=%x", WSAGetLastError());
		SendMessage(g_hwndLog, WM_USER_LOG, 0, (LPARAM) szBuffer);
		return 0;
	}
	g_hListener = socket(AF_INET, SOCK_STREAM, 0);

	if (g_hListener == INVALID_SOCKET) {
		sprintf(szBuffer, "socket error code=%x", WSAGetLastError());
		SendMessage(g_hwndLog, WM_USER_LOG, 0, (LPARAM) szBuffer);
		return 0;
	}
	struct sockaddr_in sa;

	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(g_nPort);

	if (bind(g_hListener, (struct sockaddr*) &sa, sizeof(sa)) != 0) {
		sprintf(szBuffer, "bind error code=%x", WSAGetLastError());
		SendMessage(g_hwndLog, WM_USER_LOG, 0, (LPARAM) szBuffer);
		return 0;
	}
	if (listen(g_hListener, 5) != 0) {
		sprintf(szBuffer, "listen error code=%x", WSAGetLastError());
		SendMessage(g_hwndLog, WM_USER_LOG, 0, (LPARAM) szBuffer);
		return 0;
	}
	while (true) {
		struct sockaddr_in ca;
		int clientAddressLength = sizeof(ca);
		int nLength = sizeof(sa);
		SOCKET hSocket = accept(g_hListener, (struct sockaddr*) &ca, &nLength);
		
		if (hSocket == INVALID_SOCKET) {
			continue;
		}
		sprintf(szBuffer, "new connection %d.%d.%d.%d", 
			sa.sin_addr.S_un.S_un_b.s_b1, sa.sin_addr.S_un.S_un_b.s_b2,
			sa.sin_addr.S_un.S_un_b.s_b3, sa.sin_addr.S_un.S_un_b.s_b4);
		Log(szBuffer);

		User* pUser = new User(hSocket);

		g_world.AddUser(pUser);
		_beginthreadex(NULL, 0, UserThread, pUser, 0, NULL);
		pUser->Send(S_NAME);
    }
	if (g_hListener != 0) {
		if (closesocket(g_hListener) == SOCKET_ERROR) {
			sprintf(szBuffer, "closesocket error code=%x", WSAGetLastError());
			SendMessage(g_hwndLog, WM_USER_LOG, 0, (LPARAM) szBuffer);
		}
	}
    WSACleanup();
    return 0;
}

unsigned __stdcall UserThread(void* pArg)
{
	User* pUser = (User*) pArg;
	enum MODE { PACKET_LENGTH, PACKET_BODY };
	MODE mode = PACKET_LENGTH;
	char szBuffer[512];
	short sPacketLen;
	int nPI = 0;
	int nRI = 0;
	WSAEVENT hRecvEvent = WSACreateEvent();
	int nRead;
	WSANETWORKEVENTS event;

	//  б ݱ⸦ ̺Ʈ 
	WSAEventSelect(pUser->Socket(), hRecvEvent, FD_READ | FD_CLOSE);

	while (g_bShutDown == false) {
		// ̺Ʈ ٸ
		DWORD dwReturn = WSAWaitForMultipleEvents(1, &hRecvEvent, FALSE, WSA_INFINITE, FALSE);

		if (dwReturn == WSA_WAIT_FAILED) {
			int nError = WSAGetLastError();
		}
		else if (dwReturn == 0) {
			// ̺Ʈ ȣ 
			WSAResetEvent(hRecvEvent);
		}
		while ((nRead = recv(pUser->Socket(), (char*) szBuffer + nPI, 512 - nPI, 0)) > 0) {
			nRI += nRead;

			while (nPI < nRI) {
				if (mode == PACKET_LENGTH) {
					if (nPI + 2 <= nRI) {
						// Ŷ 
						sPacketLen = szBuffer[0] + (szBuffer[1] << 8) - 2;
						nPI += 2;
						mode = PACKET_BODY;
					}
					else {
						break;
					}
				}
				else if (mode == PACKET_BODY) {
					if (nPI + sPacketLen <= nRI) {
						// Ŷȣ ´ Ȯ
						if (szBuffer[nPI] >= C_MAX) {
						}
						else {
							//  Ŷ ó
							pUser->Packet(szBuffer + nPI);
						}
						nPI += sPacketLen;
						mode = PACKET_LENGTH;
					}
				}
			}
			if (nPI == nRI) {
				nPI = nRI = 0;
			}
			else if (nPI > 0) {
				memmove(szBuffer, szBuffer + nPI, nRI - nPI);
				nRI -= nPI;
			}
		}
		// ߻ ̺Ʈ  Ȯ
		WSAEnumNetworkEvents(pUser->Socket(), hRecvEvent, &event);
		// FD_CLOSE ̺Ʈ  
		if ((event.lNetworkEvents & FD_CLOSE) == FD_CLOSE) {
			g_world.Remove(pUser);
			break;
		}
	}
	WSACloseEvent(hRecvEvent);
	return 0;
}

