#include <windows.h>
#include "server.h"
#include "world.h"
#include "util.h"
#include "user.h"
#include "protocol.h"

char g_pszMap[][1024] = {
//             1         2         3         4         5         6    
//	 0123456789012345678901234567890123456789012345678901234567890123456
	"+-----------------------------------------------------------------+", // 0
	"|                                                                 |", // 1
	"|                                                                 |", // 2
	"|                                                                 |", // 3
	"|     +---------+                             +----------+        |", // 4
	"|     |         |                             |          |        |", // 5
	"|     |         |             +-+                        |        |", // 6
	"|     |         |             | |             |          |        |", // 7
	"|     |         |             +-+             |          |        |", // 8
	"|     +---- ----+                             +----------+        |", // 9
	"|                                                                 |", // 10
	"|                                                                 |", // 11
	"|                                                                 |", // 12
	"|                                                                 |", // 13
	"|                                                                 |", // 14
	"|     +-----   ------+---------------------+-- ---------------+---+", // 15
	"|     |              |                     |                  |   |", // 16
	"|     |              |  ----------------+  +---------------   |   |", // 17
	"|     |              |                  |                     |   |", // 18
	"|     |              |  +------------+  +---------------------+   |", // 19
	"|     |              |  |                                     |   |", // 20
	"|     |              |  |               +------------------+  |   |", // 21
	"|     +-------+------+  +-----------+   |                  |  |   |", // 22
	"|             |                     |   +----------------  |  |   |", // 23
	"|             |                     |                      |      |", // 24
	"+-------------+---------------------+----------------------+------+"  // 25
};

World::World()
{
	InitializeCriticalSection(&m_csUser);
	InitializeCriticalSection(&m_csWorld);
	m_nId = 0;
}

World::~World()
{
	DeleteCriticalSection(&m_csUser);
	DeleteCriticalSection(&m_csWorld);
}

void World::Send(char* pPacket,int nPacket)
{
	std::map<unsigned int,User*>::iterator i;

	EnterCriticalSection(&m_csUser);
	for (i = m_mapUser.begin(); i != m_mapUser.end(); i++) {
		User* pTarget = i->second;
		
		pTarget->Send(pPacket, nPacket);
	}
	LeaveCriticalSection(&m_csUser);
}

bool World::MoveUser(User* pUser,char cDirection)
{
	bool bMove = false;

	EnterCriticalSection(&m_csWorld);
	switch (cDirection) {
	case UP :
		if (g_pszMap[pUser->Y() - 1][pUser->X()] == ' ') {
			g_pszMap[pUser->Y()][pUser->X()] = ' ';
			g_pszMap[pUser->Y() - 1][pUser->X()] = 'U';
			pUser->SetY(pUser->Y() - 1);
			bMove = true;
		}
		break;
	case DOWN :
		if (g_pszMap[pUser->Y() + 1][pUser->X()] == ' ') {
			g_pszMap[pUser->Y()][pUser->X()] = ' ';
			g_pszMap[pUser->Y() + 1][pUser->X()] = 'U';
			pUser->SetY(pUser->Y() + 1);
			bMove = true;
		}
		break;
	case LEFT :
		if (g_pszMap[pUser->Y()][pUser->X() - 1] == ' ') {
			g_pszMap[pUser->Y()][pUser->X()] = ' ';
			g_pszMap[pUser->Y()][pUser->X() - 1] = 'U';
			pUser->SetX(pUser->X() - 1);
			bMove = true;
		}
		break;
	case RIGHT :
		if (g_pszMap[pUser->Y()][pUser->X() + 1] == ' ') {
			g_pszMap[pUser->Y()][pUser->X()] = ' ';
			g_pszMap[pUser->Y()][pUser->X() + 1] = 'U';
			pUser->SetX(pUser->X() + 1);
			bMove = true;
		}
		break;
	};
	LeaveCriticalSection(&m_csWorld);

	return bMove;
}

void World::Move(User* pUser,char cDirection)
{
	if (MoveUser(pUser, cDirection) == false) {
		return;
	}
	char szPacket[512];
	char* pPacket = szPacket + 2;

	pPacket = SetChar(pPacket, S_MOVE);
	pPacket = SetInteger(pPacket, pUser->Id());
	pPacket = SetInteger(pPacket, pUser->X());
	pPacket = SetInteger(pPacket, pUser->Y());
	SetShort(szPacket, pPacket - szPacket);
	Send(szPacket, pPacket - szPacket);
}

void World::Say(User* pUser,char* pszSay)
{
	char szPacket[512];
	char* pPacket = szPacket + 2;

	pPacket = SetChar(pPacket, S_SAY);
	pPacket = SetInteger(pPacket, pUser->Id());
	pPacket = SetString(pPacket, pszSay);
	SetShort(szPacket, pPacket - szPacket);
	Send(szPacket, pPacket - szPacket);
}

void World::Attack(User* pUser,unsigned int nVictim)
{
	char szPacket[512];
	char* pPacket = szPacket + 2;

	pPacket = SetChar(pPacket, S_ATTACK);
	pPacket = SetInteger(pPacket, pUser->Id());
	pPacket = SetString(pPacket, pUser->Name());
	SetShort(szPacket, pPacket - szPacket);
	Send(szPacket, pPacket - szPacket);

	User* pVictim = FindUser(nVictim);

	if (pVictim != NULL) {
		pVictim->Damage(pUser);
		pVictim->Release();
	}
}

void World::Enter(User* pUser)
{
	char szPacket[512];
	char* pPacket = szPacket + 2;
	bool bPut = false;

	EnterCriticalSection(&m_csWorld);
	for (int nY = 10; nY < 25 && bPut == false; nY++) {
		for (int nX = 11; nX < 65 && bPut == false; nX++) {
			if (g_pszMap[nY][nX] == ' ') {
				pUser->SetX(nX);
				pUser->SetY(nY);
				int nNum = strlen(g_pszMap[nY]);
				g_pszMap[nY][nX] = 'U'; // [nX] = 'U';
				bPut = true;
			}
		}
	}
	LeaveCriticalSection(&m_csWorld);

	pPacket = SetChar(pPacket, S_ENTER);
	pPacket = SetInteger(pPacket, pUser->Id());
	pPacket = SetString(pPacket, pUser->Name());
	pPacket = SetInteger(pPacket, pUser->X());
	pPacket = SetInteger(pPacket, pUser->Y());
	SetShort(szPacket, pPacket - szPacket);
	Send(szPacket, pPacket - szPacket);

	std::map<unsigned int,User*>::iterator i;
	User* pTarget;

	EnterCriticalSection(&m_csUser);
	for (i = m_mapUser.begin(); i != m_mapUser.end(); i++) {
		pTarget = i->second;

		if (pTarget != pUser) {
			pPacket = szPacket + 2;
			pPacket = SetChar(pPacket, S_ENTER);
			pPacket = SetInteger(pPacket, pTarget->Id());
			pPacket = SetString(pPacket, pTarget->Name());
			pPacket = SetInteger(pPacket, pTarget->X());
			pPacket = SetInteger(pPacket, pTarget->Y());
			SetShort(szPacket, pPacket - szPacket);
			pUser->Send(szPacket, pPacket - szPacket);
		}
	}
	LeaveCriticalSection(&m_csUser);
}

void World::AddUser(User* pUser)
{
	int nId = InterlockedIncrement((long*) &m_nId);

	EnterCriticalSection(&m_csUser);
	pUser->AddRef();
	pUser->SetId(nId);
	m_mapUser.insert(std::map<unsigned int,User*>::value_type(nId, pUser));
	LeaveCriticalSection(&m_csUser);
}

User* World::FindUser(unsigned int nId)
{
	User* pUser = NULL;
	std::map<unsigned int,User*>::iterator i;

	EnterCriticalSection(&m_csUser);
	i = m_mapUser.find(nId);
	if (i != m_mapUser.end()) {
		pUser = i->second;
	}
	LeaveCriticalSection(&m_csUser);
	return pUser;
}

void World::Remove(User* pUser)
{
	std::map<unsigned int,User*>::iterator i;
	char szPacket[512];
	char* pPacket = szPacket + 2;

	pPacket = SetChar(pPacket, S_REMOVE);
	pPacket = SetInteger(pPacket, pUser->Id());
	SetShort(szPacket, pPacket - szPacket);

	EnterCriticalSection(&m_csWorld);
	g_pszMap[pUser->Y()][pUser->X()] = ' ';
	LeaveCriticalSection(&m_csWorld);

	EnterCriticalSection(&m_csUser);
	i = m_mapUser.find(pUser->Id());
	if (i != m_mapUser.end()) {
		delete i->second;
		m_mapUser.erase(i);
	}
	LeaveCriticalSection(&m_csUser);

	Send(szPacket, pPacket - szPacket);
}
